# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

import os
import re
from pathlib import Path
from typing import List, Optional, Dict, Tuple

from uncertainties import ufloat

from .fileutils import PathLike
from .fundamental_data import ATOMIC_SYMBOL, ATOMIC_NUMBER, isotopes, atomic_mass
from .plugin import PluginGeneric, _find_executable
from .results import Results


def expand_element(xsdir: Optional[PathLike] = None):
    def expand_element_inner(material: str, default_suffix: str = None) -> str:
        """Expand elements in an MCNP material definition

        Parameters
        ----------
        material
            String representing material definition
        default_suffix
            Cross section suffix used by default if none is provided

        Returns
        -------
        str
            Material definition string with elements expanded

        """
        lines_in = material.split('\n')

        # Note that the 'xsdir' variable is in an enclosing scope -- it gets
        # passed in by PluginMCNP so that each instance of the plugin can
        # uniquely set its own xsdir
        available_nuclides = _get_nuclides_from_xsdir(xsdir)

        lines_out = []
        for line in lines_in:
            # If the line is a comment, don't modify it
            if re.match('^[ ]*[Cc] ', line):
                lines_out.append(line)
                continue

            # Ignore end-of-line comments
            index_comment = line.find('$')
            if index_comment != -1:
                line = line[:index_comment]

            words = line.split()
            i = 0
            while i < len(words):
                # Determine whether line includes start of material card
                if re.match(r'[Mm]\d+', words[i]):
                    start = f'{words[i]:<4} '
                    i += 1
                else:
                    start = '     '

                zaid_with_suffix = words[i]
                conc = words[i + 1]
                i += 2

                # Determine ZAID and suffix used
                zaid, *original_suffix = zaid_with_suffix.split('.')

                # If no '.' is present, original suffix is empty. If '.' is present
                # but no suffix afterward, original_surffix has a single empty
                # string. For either of these cases, use the default suffix
                no_suffix = (not original_suffix or not original_suffix[0])
                if no_suffix and default_suffix is not None:
                    suffix = default_suffix
                else:
                    suffix = original_suffix[0]

                # Determine Z and A
                if zaid.isalpha():
                    Z = ATOMIC_NUMBER[zaid]
                    A = 0
                else:
                    Z, A = divmod(int(zaid), 1000)

                # Split into isotopes if natural element is given
                if A == 0:
                    conc = float(conc)
                    weight_percent = conc < 0
                    symbol = ATOMIC_SYMBOL[Z]

                    # Determine what isotopes to add
                    available_isotopes = available_nuclides.get((Z, suffix), [])
                    natural_isotope_fractions = isotopes(symbol)
                    natural_isotopes = [isotope for isotope, _ in natural_isotope_fractions]
                    missing_isotopes = set(natural_isotopes) - set(available_isotopes)
                    if len(natural_isotopes) == 0:
                        # No natural isotopes, can not expand
                        raise ValueError(
                            f"{zaid_with_suffix} cannot be expanded because it has "
                            "no naturally occurring isotopes.")
                    if len(available_isotopes) == 1:
                        # Case 1 -- only a single isotope available. In this case, all
                        # the concentration goes to that single isotope
                        isotope_fractions = [(available_isotopes[0], 1.0)]
                    elif len(missing_isotopes) == 0:
                        # Case 2 -- all isotopes are available, use as is!
                        isotope_fractions = natural_isotope_fractions
                    elif available_isotopes == ['C0', 'C13']:
                        # Case 3 -- special case in JEFF 3.3 where both elemental C
                        # and isotopic C13 are available
                        isotope_fractions = [('C0', 1.0)]
                    elif len(missing_isotopes) == 1:
                        # Case 4 -- one missing isotope. In this case, lump it into
                        # whatever natural isotope has the highest abundance

                        # Determine which isotope has highest abundance
                        highest_item = max(natural_isotope_fractions, key=lambda x: x[1])
                        highest_index = natural_isotope_fractions.index(highest_item)

                        # Determine index/fraction of missing isotope
                        missing_item, = missing_isotopes
                        missing_index = natural_isotopes.index(missing_item)
                        missing_fraction = natural_isotope_fractions[missing_index][1]

                        # Replace missing isotope with highest abundance one
                        natural_isotope_fractions[highest_index] = (
                            highest_item[0], highest_item[1] + missing_fraction)
                        natural_isotope_fractions.pop(missing_index)
                        isotope_fractions = natural_isotope_fractions
                    else:
                        raise ValueError(
                            f"Could not expand {zaid}; no corresponding isotopes "
                            f"found in xsdir file.")

                    if weight_percent:
                        # Convert molar fractions to mass fractions
                        total = sum([fraction*atomic_mass(isotope)
                                     for isotope, fraction in isotope_fractions])
                        isotope_fractions = [
                            (isotope, fraction/total*atomic_mass(isotope))
                            for isotope, fraction in isotope_fractions
                        ]

                    for isotope, fraction in isotope_fractions:
                        iso_A = int(*re.match(rf'{symbol}(\d+)', isotope).groups())
                        lines_out.append(f"{start}{Z}{iso_A:03}.{suffix} {conc * fraction}")

                        # Prevent 'm#' from being written multiple times
                        start = '     '
                else:
                    lines_out.append(f"{start}{zaid_with_suffix} {conc}")

        return "\n".join(lines_out)

    return expand_element_inner


def _get_nuclides_from_xsdir(path: Optional[PathLike] = None) -> Dict[Tuple[int, str], List[str]]:
    """Determine available nuclides from an MCNP xsdir file.

    Parameters
    ----------
    path
        Path to xsdir file

    Returns
    -------
    dict
        Dictionary mapping (Z, suffix) to a list of available nuclides
    """
    if path is None:
        datapath = os.environ.get('DATAPATH')
        if datapath is None:
            raise EnvironmentError(
                "Need to set DATAPATH environment vairable to determine what "
                "MCNP cross section libraries are available.")
        path = Path(datapath) / 'xsdir'

    # Find 'directory' section
    with open(path, 'r') as fh:
        lines = fh.readlines()
    for index, line in enumerate(lines):
        if line.strip().lower() == 'directory':
            break
    else:
        raise RuntimeError("Could not find 'directory' section in MCNP xsdir file")

    # Handle continuation lines indicated by '+' at end of line
    lines = lines[index + 1:]
    continue_lines = [i for i, line in enumerate(lines)
                      if line.strip().endswith('+')]
    for i in reversed(continue_lines):
        lines[i] = lines[i].strip()[:-1] + lines.pop(i + 1)

    # Create list of ACE libraries
    tables = {}
    for line in lines:
        words = line.split()
        if len(words) < 3:
            continue

        if not words[0].endswith('c'):
            continue

        zaid, suffix = words[0].split('.')
        Z, A = divmod(int(zaid), 1000)
        symbol = ATOMIC_SYMBOL[Z]
        if (Z, suffix) not in tables:
            tables[Z, suffix] = []
        tables[Z, suffix].append(f'{symbol}{A}')

    return tables


class ResultsMCNP(Results):
    """MCNP simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    input_file
        Rendered MCNP input file
    keff
        K-effective value
    stdout
        Standard output from MCNP run
    """

    @property
    def input_file(self) -> str:
        return self.inputs[0].read_text()

    @property
    def keff(self) -> ufloat:
        with open(self.base_path / 'outp', 'r') as f:
            for line in f:
                if line.strip().startswith('col/abs/trk len'):
                    words = line.split()
                    mean = float(words[2])
                    stdev = float(words[3])
                    return ufloat(mean, stdev)
            else:
                raise ValueError(
                    "Could not determine final k-effective value from MCNP output")


class PluginMCNP(PluginGeneric):
    """Plugin for running MCNP

    In addition to the basic capability to use placeholders in MCNP input files,
    this class also provides a custom Jinja filter called ``expand_element``
    that allows you to specify natural elements in MCNP material definitions and
    have them automatically expanded based on what isotopes appear in the xsdir
    file.

    Parameters
    ----------
    template_file
        Templated MCNP input
    executable
        Path to MCNP executable
    xsdir
        Path to the xsdir file used for natural element expansion. Defaults to
        the file named 'xsdir' under the directory specified by the
        :envvar:`DATAPATH` environment variable.
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when MCNP is run
    show_stderr
        Whether to display output from stderr when MCNP is run

    Attributes
    ----------
    executable
        Path to MCNP executable
    execute_command
        List of command-line arguments used to call the executable

    """

    def __init__(
        self,
        template_file: str,
        executable: PathLike = 'mcnp6',
        xsdir: Optional[PathLike] = None,
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'MCNP_DIR')
        super().__init__(
            executable, ['{self.executable}', 'i={self.input_name}'],
            template_file, extra_inputs, extra_template_inputs, "MCNP",
            show_stdout, show_stderr, unit_system='cgs')
        self.input_name = "mcnp_input"

        # Add custom 'expand_element' Jinja filter
        self.render_template.environment.filters['expand_element'] = expand_element(xsdir)
        for renderer in self.extra_render_templates:
            renderer.environment.filters['expand_element'] = expand_element(xsdir)



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from collections.abc import Sequence
from pathlib import Path
import pprint
import shutil
from typing import Union
from warnings import warn

import platformdirs

from .results import Results


class Database(Sequence):
    """Database of simulation results

    Parameters
    ----------
    path
        Path to database directory

    Attributes
    ----------
    default_path
        Path used by default when creating instances if no path is specified
    job_id
        Integer ID assigned to new results
    path
        Base path for the database directory
    results
        List of simulation results in database

    """

    _default_path = platformdirs.user_data_path('watts')
    _instances = {}

    def __new__(cls, path=None):
        # If no path specified, use global default
        if path is None:
            path = cls._default_path

        # If this class has already been instantiated before, return the
        # corresponding instance
        abs_path = Path(path).resolve()
        if abs_path in Database._instances:
            return Database._instances[abs_path]
        else:
            return super().__new__(cls)

    def __init__(self, path=None):
        # If no path specified, use global default
        if path is None:
            path = self._default_path

        # If instance has already been created, no need to perform setup logic
        if path in Database._instances:
            return

        # Create database directory if it doesn't already
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        self._path = path

        # Read previous results
        self._results = []
        for dir in sorted(self.path.iterdir(), key=lambda x: x.stat().st_ctime):
            try:
                self._results.append(Results.from_pickle(dir / ".result_info.pkl"))
            except Exception:
                warn(f"Could not read results from {dir}")

        # Determine unique job ID based on what has already been used
        used_job_ids = set()
        for result in self:
            job_id = getattr(result, 'job_id', None)
            if job_id is not None:
                used_job_ids.add(job_id)
        self.job_id = max(used_job_ids, default=-1) + 1

        # Add instance to class-wide dictionary
        Database._instances[path.resolve()] = self

    def __repr__(self):
        return pprint.pformat(self._results)

    def __getitem__(self, index):
        return self._results[index]

    def __len__(self):
        return len(self._results)

    @property
    def path(self) -> Path:
        return self._path

    @property
    def default_path(self) -> Path:
        return self.get_default_path()

    @default_path.setter
    def default_path(self, path):
        self.set_default_path(path)

    @classmethod
    def set_default_path(cls, path: Union[str, Path]):
        """Set the default path used when instances are created

        Parameters
        ----------
        path
            Default path to use
        """

        cls._default_path = Path(path).resolve()

    @classmethod
    def get_default_path(cls) -> Path:
        """Get the default path used when instances are created

        Returns
        -------
        Default path
        """
        return cls._default_path

    def add_result(self, result: Results):
        """Add a result to the database

        Parameters
        ----------
        result
            Simulation results to add

        """
        self._results.append(result)

        # Save result info that can be recreated
        result.save(result.base_path / ".result_info.pkl")

    def clear(self):
        """Remove all results from database"""
        for dir in self.path.iterdir():
            shutil.rmtree(dir)
        self._results.clear()

    def remove(self, result: Results):
        """Remove a single result from the database

        Parameters
        ----------
        result
            Result to remove from the database

        """
        self._results.remove(result)
        shutil.rmtree(result.base_path)

    def show_summary(self):
        """Show a summary of results in database"""
        for result in self._results:
            rel_path = result.base_path.relative_to(self.path)
            print(result.time, result.plugin, str(rel_path),
                  f"({len(result.inputs)} inputs)",
                  f"({len(result.outputs)} outputs)")



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from pathlib import Path
import sys
from typing import List, Optional

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import PluginGeneric, _find_executable
from .results import Results, ExecInfo


class PluginABCE(PluginGeneric):
    """Plugin for running ABCE

    Parameters
    ----------
    template_file
        Templated ABCE input
    executable
        Path to ABCE run.py script
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when ABCE is run
    show_stderr
        Whether to display output from stderr when ABCE is run

    Attributes
    ----------
    executable
        Path to ABCE executable
    execute_command
        List of command-line arguments used to call the executable

    """


    def __init__(
        self,
        template_file: str,
        executable: PathLike = 'run.py',
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'ABCE_DIR')
        execute_command = [sys.executable, '{self.executable}', '--settings_file', '{self.input_name}']
        super().__init__(
            executable, execute_command, template_file, extra_inputs,
            extra_template_inputs, 'ABCE', show_stdout, show_stderr)
        self.input_name = 'settings.yml'

    @PluginGeneric.executable.setter
    def executable(self, exe: PathLike):
        if not exe.is_file():
            raise RuntimeError(
                f"{self.plugin_name} script '{exe}' does not exist. The "
                "ABCE_DIR environment variable needs to be set to a directory "
                "containing the run.py script."
            )
        self._executable = Path(exe)


class ResultsABCE(Results):
    """ABCE simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from ABCE run
    """
    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        super().__init__(params, exec_info, inputs, outputs)



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

import sys

import click
from prettytable import PrettyTable

from . import Database


@click.group()
def main():
    pass


@click.command()
@click.option('--job-id', default=None, type=int, help='Filter by job ID')
@click.option('--last-job', is_flag=True, help='Display most recent job')
@click.option('--plugin', default=None, help='Filter by plugin name')
@click.option('--name', default=None, help='Filter by name')
@click.option('--database', default=None, help='Path to database')
def results(job_id, last_job, plugin, name, database):
    """List results"""
    db = Database(database) if database else Database()
    table = PrettyTable(field_names=['Index', 'Job ID', 'Plugin', 'Name', 'Time'], align='l')

    # Determine most recent job
    if last_job:
        job_ids = set()
        for result in db:
            if result.job_id is not None:
                job_ids.add(result.job_id)
        if job_ids:
            job_id = max(job_ids)

    for i, result in enumerate(db):
        # Apply filters
        if job_id is not None and job_id != result.job_id:
            continue
        if plugin is not None and plugin != result.plugin:
            continue
        if name is not None and name != result.name:
            continue

        table.add_row([i, result.job_id, result.plugin, result.name, result.time])
    click.echo(table.get_string())


@click.command()
@click.option('--database', default=None, help='Path to database')
@click.argument('index', type=int)
def dir(database, index):
    """Show directory containing files for a specific result

    The 'index' can be determined from the Index column when running 'watts
    results'.

    """
    db = Database(database) if database else Database()
    try:
        result = db[index]
    except IndexError:
        click.echo(f"No result with index {index} in database at {db.path}", err=True)
        sys.exit(1)
    click.echo(result.base_path)


@click.command()
@click.option('--database', default=None, help='Path to database')
@click.argument('index', type=int)
def stdout(database, index):
    """Show standard output from a specific result

    The 'index' can be determined from the Index column when running 'watts
    results'.

    """
    db = Database(database) if database else Database()
    try:
        result = db[index]
    except IndexError:
        click.echo(f"No result with index {index} in database at {db.path}", err=True)
        sys.exit(1)
    click.echo(result.stdout)


@click.command()
@click.option('--database', default=None, help='Path to database')
@click.option('--all', is_flag=True, help='Remove all results')
@click.argument('index', type=int, required=False)
def rm(database, all, index):
    """Remove a result from the database

    The 'index' can be determined from the Index column when running 'watts
    results'. Note that removing a result will delete all input and output files
    associated with the result

    """
    db = Database(database) if database else Database()

    if not all and index is None:
        click.echo("Must either provide index argument or --all flag.", err=True)
        sys.exit(1)

    if all:
        db.clear()
    else:
        try:
            result = db[index]
        except IndexError:
            click.echo(f"No result with index {index} in database at {db.path}", err=True)
            sys.exit(1)
        db.remove(result)


main.add_command(results)
main.add_command(dir)
main.add_command(stdout)
main.add_command(rm)



# SPDX-FileCopyrightText: 2024 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from pathlib import Path
import shutil
import subprocess
from typing import List, Optional
import os
import pandas as pd

from .plugin import Plugin
from .results import Results, ExecInfo
from .fileutils import PathLike
from .parameters import Parameters
from .template import TemplateRenderer


class ResultsGCMAT(Results):
    """GCMAT simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from GCMAT run
    csv_data
        Data from the output CSV file
    """
    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        super().__init__(params, exec_info, inputs, outputs)
        self.csv_data = self._get_gcmat_csv_data()

    def _get_gcmat_csv_data(self) -> pd.DataFrame:
        """Read GCMAT output CSV file and return results as a DataFrame"""
        output_file = next((p for p in self.outputs if p.name == 'GUIOutputs.csv'), None)
        if output_file and output_file.exists():
            return pd.read_csv(output_file)
        else:
            return pd.DataFrame()  # Return an empty DataFrame if no CSV file is found


class PluginGCMAT(Plugin):
    """Plugin for running GCMAT

    Parameters
    ----------
    template_file
        Template file used to generate the input files
    extra_inputs
        Extra (non-templated) input files
    show_stdout
        Whether to display output from stdout when GCMAT is run
    show_stderr
        Whether to display output from stderr when GCMAT is run

    """
    def __init__(self, template_file: PathLike,
                 extra_inputs: Optional[List[PathLike]] = None,
                 show_stdout: bool = False, show_stderr: bool = False):
        super().__init__(extra_inputs, show_stdout, show_stderr)
        self.template_file = template_file
        self.plugin_name = 'GCMAT'
        self.renderer = TemplateRenderer(template_file)
        self.gcmat_dir = os.getenv('GCMAT_DIR')
        if not self.gcmat_dir:
            raise EnvironmentError("GCMAT_DIR environment variable is not set.")

        # Include './run_repast.sh' as the executable and all files in the 'data' folder as default extra inputs
        self.executable = Path(self.gcmat_dir) / "run_repast.sh"
        self.default_extra_inputs = list((Path(self.gcmat_dir) / "complete_model" / "data").glob("**/*"))

        # Initialize output_folder attribute
        self.output_folder = None

    def prerun(self, params: Parameters) -> None:
        """Generate GCMAT input files

        Parameters
        ----------
        params
            Parameters used by the GCMAT template
        """
        # Render the template to create the input file
        input_file = Path("gc_input.txt")
        self.renderer(params, filename=input_file)

        # Copy the input file to the required directory
        model_directory = Path(self.gcmat_dir) / "complete_model"
        target_directory = model_directory / "data/scenariosNuclear/default_UserInputs"
        target_directory.mkdir(parents=True, exist_ok=True)
        shutil.copy(input_file, target_directory / "demandScenarioV2.txt")

    def run(self, end_year: int = 2080, output_folder: str = "testout", **kwargs):
        """Run GCMAT

        Parameters
        ----------
        end_year
            The year to end the simulation
        output_folder
            The folder where outputs will be stored
        kwargs
            Additional keyword arguments to pass to the subprocess
        """
        # use the absolute path for the output folder
        self.output_folder = os.path.join(self.gcmat_dir, output_folder)
        param_string = f'1\tendAt\t{end_year}'
        command = [str(self.executable), param_string, subprocess.check_output('realpath .', shell=True).strip().decode('utf-8'), output_folder]
        # Run the GCMAT simulation
        subprocess.run(command, cwd=self.gcmat_dir, **kwargs)

    def postrun(self, params: Parameters, exec_info: ExecInfo) -> ResultsGCMAT:
        """Collect information from GCMAT simulation and create results object

        Parameters
        ----------
        params
            Parameters used to create GCMAT model
        exec_info
            Execution information

        Returns
        -------
        GCMAT results object
        """
        output_folder = Path(self.output_folder)  # Retrieve the stored
        # Only collect the GUIOutputs.csv file
        # can add more files if needed
        outputs = []
        gui_outputs_file = output_folder / "GUIOutputs.csv"
        if gui_outputs_file.exists():
            outputs.append(gui_outputs_file)
        return ResultsGCMAT(params, exec_info, self.extra_inputs, outputs)



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod
from contextlib import redirect_stdout, redirect_stderr
import os
from pathlib import Path
import shutil
import time
from typing import Optional, List, Union
import uuid

from .database import Database
from .fileutils import cd_tmpdir, PathLike, tee_stdout, tee_stderr, run as run_proc
from .parameters import Parameters
from .results import Results, ExecInfo
from .template import TemplateRenderer
import watts


class Plugin(ABC):
    """Class defining the Plugin interface

    Parameters
    ----------
    extra_inputs
        Extra (non-templated) input files
    show_stdout
        Whether to display output from stdout when :math:`run` is called
    show_stderr
        Whether to display output from stderr when :meth:`run` is called
    unit_system : {'si', 'cgs'}
        Unit system to convert to when rendering input files

    Attributes
    ----------
    plugin_name : str
        Name of the plugin
    unit_system : {'si', 'cgs'}
        Desired system of units for rendering templates

    """

    def __init__(
        self,
        extra_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False,
        unit_system: str = 'si'
    ):
        self.extra_inputs = []
        if extra_inputs is not None:
            self.extra_inputs = [Path(f).resolve() for f in extra_inputs]
        self.show_stdout = show_stdout
        self.show_stderr = show_stderr
        self.unit_system = unit_system

    @abstractmethod
    def prerun(self, params: Parameters):
        ...

    @abstractmethod
    def run(self):
        ...

    @abstractmethod
    def postrun(self, params: Parameters, exec_info: ExecInfo) -> Results:
        ...

    def __call__(
        self,
        params: Parameters = None,
        name: str = '',
        output_dir: Optional[PathLike] = None,
        verbose: bool = True,
        cleanup: bool = True,
        **kwargs
    ) -> Results:
        """Run the complete workflow for the plugin

        Parameters
        ----------
        params
            Parameters used in generating inputs
        name
            Name associated with execution of plugin
        output_dir
            Relative path of the directory created in the database. If None, a
            unique directory name is created using the uuid module.
        verbose
            Whether to print execution information
        cleanup
            Determines whether the temporary directory will be cleaned up
            immediately after execution.
        **kwargs
            Keyword arguments passed to the `run` method

        Returns
        -------
        Results from running plugin
        """
        db = Database()
        plugin_name = self.plugin_name
        if verbose:
            if name:
                print(f'[watts] Calling {plugin_name} ({name})...')
            else:
                print(f'[watts] Calling {plugin_name}...')

        # Generate empty Parameters object if none provided
        if params is None:
            params = Parameters()

        # Create execution info
        timestamp = time.time_ns()
        exec_info = ExecInfo(db.job_id, plugin_name, name, timestamp)

        with cd_tmpdir(cleanup=cleanup):
            # Copy extra inputs to execution directory
            cwd = Path.cwd()
            for path in self.extra_inputs:
                shutil.copy(str(path), str(cwd))  # Remove str() for Python 3.8+

            # Generate input files and perform any other prerun actions
            self.prerun(params)

            # Execute the code, redirecting stdout/stderr if requested
            with open(f'{plugin_name}_log.txt', 'w') as outfile:
                func_stdout = tee_stdout if self.show_stdout else redirect_stdout
                func_stderr = tee_stderr if self.show_stderr else redirect_stderr
                with func_stdout(outfile), func_stderr(outfile):
                    self.run(**kwargs)

            # Collect results and perform any postrun actions
            result = self.postrun(params, exec_info)

            # Create new directory for results and move files there
            if output_dir is None:
                output_dir = uuid.uuid4().hex
            workflow_path = db.path / output_dir
            workflow_path.mkdir()
            try:
                result.move_files(workflow_path)
            except Exception:
                # If error occurred, make sure we remove results directory so it
                # doesn't pollute database
                #shutil.rmtree(workflow_path)
                raise

        # Add result to database
        db.add_result(result)

        return result


class PluginGeneric(Plugin):
    """Plugin that relies on generating a template file

    This class can be used to control the execution of an arbitrary executable,
    first rendering one or more templated input files.

    Parameters
    ----------
    executable
        Path to executable
    execute_command
        List of command-line arguments, where each is formatted using the
        instance of the class as ``self``. The first string normally indicates
        the executable, i.e. "{self.executable}". The rendered input file can be
        accessed as "{self.input_name}". A single string of command-line
        arguments is also accepted.
    template_file
        Path to template file
    extra_inputs
        Extra (non-templated) input files
    extra_template_inputs
        Extra templated input files
    plugin_name
        Name of the plugin
    show_stdout
        Whether to display output from stdout when :math:`run` is called
    show_stderr
        Whether to display output from stderr when :meth:`run` is called
    unit_system : {'si', 'cgs'}
        Unit system to convert to when rendering input files

    Attributes
    ----------
    executable
        Path to plugin executable
    execute_command
        List of command-line arguments used to call the executable

    """
    def __init__(self,
        executable: PathLike,
        execute_command: Union[List[str], str],
        template_file: PathLike,
        extra_inputs: Optional[List[PathLike]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        plugin_name: str = 'Generic',
        show_stdout: bool = False,
        show_stderr: bool = False,
        unit_system: str = 'si',
    ):
        super().__init__(extra_inputs, show_stdout, show_stderr, unit_system)
        self.plugin_name = plugin_name
        self.render_template = TemplateRenderer(template_file)
        self.extra_render_templates = []
        self.input_name = 'input_rendered'
        if extra_template_inputs is not None:
            self.extra_render_templates = [TemplateRenderer(f, '') for f in extra_template_inputs]

        self.executable = executable
        if isinstance(execute_command, str):
            self._execute_command = execute_command.split()
        else:
            self._execute_command = execute_command

    @property
    def executable(self) -> Path:
        return self._executable

    @executable.setter
    def executable(self, exe: PathLike):
        if shutil.which(exe) is None:
            raise RuntimeError(f"{self.plugin_name} executable '{exe}' is missing.")
        self._executable = Path(exe)

    @property
    def execute_command(self) -> List[str]:
        return [item.format(self=self) for item in self._execute_command]

    def prerun(self, params: Parameters, filename: Optional[str] = None):
        """Render the template based on model parameters

        Parameters
        ----------
        params
            Parameters used to render template
        filename
            Filename for rendered template
        """
        # If the 'input_name' attribute is set, use that as default when
        # filename is not explicitly passed
        if filename is None and self.input_name is not None:
            filename = self.input_name

        # Make a copy of params and convert units if necessary -- the original
        # params remains unchanged
        params_copy = params.convert_units(system=self.unit_system)

        # Render the template
        self.render_template(params_copy, filename=filename)
        for render_template in self.extra_render_templates:
            render_template(params_copy)

    def postrun(self, params: Parameters, exec_info: ExecInfo, **kwargs) -> Results:
        """Read simulation results and create results object

        Parameters
        ----------
        params
            Parameters used to generate input files
        exec_info
            Execution information
        **kwargs
            Keyword arguments for Results subclasses

        Returns
        -------
        Results object
        """

        # Determine inputs and outputs
        inputs = [self.input_name] + [p.name for p in self.extra_inputs]
        for renderer in self.extra_render_templates:
            inputs.append(renderer.template_file.name)
        outputs = [p for p in Path.cwd().iterdir() if p.name not in inputs]

        # Get correct Results subclass and return instance
        results_cls = getattr(watts, f'Results{self.plugin_name}', Results)
        return results_cls(params, exec_info, inputs, outputs, **kwargs)

    def run(self, mpi_args: Optional[List[str]] = None,
            extra_args: Optional[List[str]] = None):
        """Run plugin

        Parameters
        ----------
        mpi_args
            MPI execute command and any additional MPI arguments to pass,
            e.g. ['mpiexec', '-n', '8'].
        extra_args
            Additional command-line arguments to append after the main command

        """
        if mpi_args is None:
            mpi_args = []
        if extra_args is None:
            extra_args = []
        run_proc(mpi_args + self.execute_command + extra_args)


def _find_executable(path: PathLike, environment_variable: str) -> Path:
    """Determine executable for a given code with a hint from environment variable

    Parameters
    ----------
    path
        Name of exeuctable or absolute path
    environment_variable
        Environment variable indicating directory where executable is found

    Returns
    -------
    Path to executable
    """

    exe = Path(path)

    # If path is already an absolute path, use it
    if exe.is_absolute():
        return exe

    # Check for environment variable
    base_dir = os.environ.get(environment_variable)
    return base_dir / exe if base_dir is not None else exe



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from pathlib import Path
import os
import sys
import tempfile
from typing import Mapping, List, Optional

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import PluginGeneric, _find_executable
from .results import Results, ExecInfo


class ResultsPyARC(Results):
    """PyARC simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files
    results_data
        PyARC results

    """

    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[Path], outputs: List[Path], results_data: dict):
        super().__init__(params, exec_info, inputs, outputs)
        self.results_data = results_data


class PluginPyARC(PluginGeneric):
    """Plugin for running PyARC

    Parameters
    ----------
    template_file
        Templated PyARC input
    executable
        Path to PyARC.py script
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when PyARC is run
    show_stderr
        Whether to display output from stderr when PyARC is run

    Attributes
    ----------
    executable
        Path to PyARC executable

    """

    def __init__(
        self,
        template_file: str,
        executable: PathLike = 'PyARC.py',
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'PyARC_DIR')
        super().__init__(executable, None, template_file, extra_inputs,
                         extra_template_inputs, "PyARC", show_stdout, show_stderr)
        self.input_name = "pyarc_input.son"

    @PluginGeneric.executable.setter
    def executable(self, exe: PathLike):
        if not exe.is_file():
            raise RuntimeError(
                f"{self.plugin_name} module '{exe}' does not exist. The "
                "PyARC_DIR environment variable needs to be set to a directory "
                "containing the PyARC.py module."
            )
        self._executable = Path(exe)

    def run(self, **kwargs: Mapping):
        """Run PyARC

        Parameters
        ----------
        **kwargs
            Keyword arguments passed on to :func:`pyarc.execute`
        """
        sys.path.insert(0, f'{self.executable.parent}')
        import PyARC
        self.pyarc = PyARC.PyARC()
        self.pyarc.user_object.do_run = True
        self.pyarc.user_object.do_postrun = True
        od = Path.cwd()

        with tempfile.TemporaryDirectory() as tmpdir:
            self.pyarc.execute(["-i", self.input_name, "-w", tmpdir, "-o", str(od)], **kwargs)
        sys.path.pop(0)  # Restore sys.path to original state
        os.chdir(od)  # TODO: I don't know why but I keep going to self.executable after execution - this is very wierd!

    def postrun(self, params: Parameters, exec_info: ExecInfo) -> ResultsPyARC:
        """Collect information from PyARC and create results object

        Parameters
        ----------
        params
            Parameters used to create PyARC model
        exec_info
            Execution information

        Returns
        -------
        PyARC results object
        """
        return super().postrun(params, exec_info, results_data=self.pyarc.user_object.results)



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from contextlib import contextmanager
import errno
import os
import platform
import select
import shutil
import subprocess
import sys
import tempfile
from typing import Union

if sys.platform != 'win32':
    import fcntl

# Type for arguments that accept file paths
PathLike = Union[str, bytes, os.PathLike]


@contextmanager
def cd_tmpdir(cleanup: bool = True):
    """Context manager to change to/return from a tmpdir.

    Parameters
    ----------
    cleanup
        Whether to clean up the temporary directory
    """
    tmpdir = tempfile.mkdtemp()
    cwd = os.getcwd()
    try:
        os.chdir(tmpdir)
        yield
    finally:
        os.chdir(cwd)
        if cleanup:
            shutil.rmtree(tmpdir)


def open_file(path: PathLike):
    """Open a file in explorer/finder/nautilus

    Parameters
    ----------
    path
        Path to directory
    """
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", str(path)])
    else:
        subprocess.Popen(["xdg-open", str(path)])


class _TeeStream:
    """This class acts as both a context manager (__enter__ and __exit__
    methods) as well as a file stream (write, flush, and isatty methods).

    Roughly based on code in https://github.com/algrebe/python-tee as well as
    the Python standard library
    """
    _stream = None

    def __init__(self, new_target):
        self._new_target = new_target
        self._old_target = None

    def write(self, message):
        self._new_target.write(message)
        self._old_target.write(message)

    def flush(self):
        self._new_target.flush()
        self._old_target.flush()

    def __enter__(self):
        self._old_target = getattr(sys, self._stream)
        setattr(sys, self._stream, self)

    def __exit__(self, exc_type, exc_inst, exc_tb):
        setattr(sys, self._stream, self._old_target)

    def isatty(self):
        return self._new_target.isatty()


class tee_stdout(_TeeStream):
    """Context manager for simulataneously writing to stdout and another stream"""
    _stream = "stdout"


class tee_stderr(_TeeStream):
    """Context manager for simulataneously writing to stderr and another stream"""
    _stream = "stderr"


def run(args):
    """Function that mimics subprocess.run but actually writes to sys.stdout and
    sys.stderr (not the same as the underlying file descriptors)

    Based on https://stackoverflow.com/a/12272262 and
    https://stackoverflow.com/a/7730201
    """
    # Windows doesn't support select.select and fcntl module so just default to
    # using subprocess.run. In this case, show_stdout/show_stderr won't work.
    if sys.platform == 'win32':
        subprocess.run(args)
        return

    # Helper function to add the O_NONBLOCK flag to a file descriptor
    def make_async(fd):
        fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)

    # Helper function to read some data from a file descriptor, ignoring EAGAIN errors
    def read_async(fd):
        try:
            return fd.read()
        except IOError as e:
            if e.errno != errno.EAGAIN:
                raise e
            else:
                return ''

    p = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    make_async(p.stdout)
    make_async(p.stderr)

    while True:
        select.select([p.stdout, p.stderr], [], [], 0)

        stdout_data = read_async(p.stdout)
        stderr_data = read_async(p.stderr)
        if stdout_data:
            sys.stdout.write(stdout_data.decode())
        if stderr_data:
            sys.stderr.write(stderr_data.decode())

        if p.poll() is not None:
            break



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from typing import List, Optional

from .fileutils import PathLike
from .plugin import PluginGeneric, _find_executable
from .results import Results


class ResultsSerpent(Results):
    """Serpent simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from Serpent run
    """


class PluginSerpent(PluginGeneric):
    """Plugin for running Serpent

    Parameters
    ----------
    template_file
        Templated Serpent input
    executable
        Path to Serpent executable
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when Serpent is run
    show_stderr
        Whether to display output from stderr when Serpent is run

    Attributes
    ----------
    executable
        Path to Serpent executable
    execute_command
        List of command-line arguments used to call the executable

    """

    def __init__(
        self,
        template_file: str,
        executable: PathLike = 'sss2',
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'SERPENT_DIR')
        super().__init__(
            executable, ['{self.executable}', '{self.input_name}'],
            template_file, extra_inputs, extra_template_inputs, "Serpent",
            show_stdout, show_stderr, unit_system='cgs'
        )
        self.input_name = "serpent_input"



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Optional

import jinja2

from .fileutils import PathLike
from .parameters import Parameters


class TemplateRenderer:
    """Helper class for rendering a Jinja template

    Parameters
    ----------
    template_file
        Path to template file
    **environment_kwargs
        Keywork arguments passed to :class:`jinja2.Environment`

    Attributes
    ----------
    environment : jinja2.Environment
        Jinja environment used to manage templates
    template_file : pathlib.Path
        Path to template file to render
    suffix : str
        Suffix added to filename when rendering a template

    """
    def __init__(self, template_file: PathLike, suffix: str = '.rendered', **environment_kwargs):
        self.environment = jinja2.Environment(
            undefined=jinja2.StrictUndefined,
            **environment_kwargs
        )
        self.template_file = Path(template_file).resolve()
        self.suffix = suffix

    def __call__(self, params: Parameters, filename: Optional[PathLike] = None):
        """Render the template

        Parameters
        ----------
        params
            User parameters used to fill placeholders
        filename
            Filename for rendered template (If none is provided, by default the
            filename of the template is changed so that the new extension is
            .rendered)
        """
        # Default rendered template filename
        if filename is None:
            name = self.template_file.name
            out_path = Path(f'{name}{self.suffix}')
        else:
            out_path = Path(filename)

        # Render template and write to file
        template = self.environment.from_string(self.template_file.read_text())
        out_path.write_text(template.render(**params))



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from pathlib import Path
import sys
from typing import List, Optional

import pandas as pd

from .fileutils import PathLike
from .plugin import PluginGeneric, _find_executable
from .results import Results


class PluginACCERT(PluginGeneric):
    """Plugin for running ACCERT

    Parameters
    ----------
    template_file
        Templated ACCERT input
    executable
        Path to ACCERT Main.py script
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when ACCERT is run
    show_stderr
        Whether to display output from stderr when ACCERT is run

    Attributes
    ----------
    executable
        Path to ACCERT executable

    """

    def __init__(self, template_file: str,
                 executable: PathLike = 'Main.py',
                 extra_inputs: Optional[List[str]] = None,
                 extra_template_inputs: Optional[List[PathLike]] = None,
                 show_stdout: bool = False, show_stderr: bool = False):
        executable = _find_executable(executable, 'ACCERT_DIR')
        execute_command = [sys.executable, '{self.executable}', '-i', '{self.input_name}']
        super().__init__(executable, execute_command, template_file, extra_inputs,
                         extra_template_inputs, "ACCERT", show_stdout, show_stderr)
        self.input_name = "ACCERT_input.son"

    @PluginGeneric.executable.setter
    def executable(self, exe: PathLike):
        if not exe.is_file():
            raise RuntimeError(
                f"{self.plugin_name} module '{exe}' does not exist. The "
                "ACCERT_DIR environment variable needs to be set to a directory "
                "containing the Main.py module."
            )
        self._executable = Path(exe)


class ResultsACCERT(Results):
    """ACCERT simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    total_cost
        ACCERT results of total cost
    account_table
        ACCERT results of account table
    """

    @property
    def total_cost(self) -> float:
        return self.account_table['total_cost'].values[0]

    @property
    def account_table(self) -> pd.DataFrame:
        account_file = self.base_path / 'ACCERT_updated_account.xlsx'
        if Path(account_file).exists():
            return pd.read_excel(account_file)
        else:
            raise FileNotFoundError('ACCERT output file not found')



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from .plugin import *
from .plugin_accert import *
from .plugin_openmc import *
from .plugin_moose import *
from .plugin_pyarc import *
from .plugin_sas import *
from .plugin_relap5 import *
from .plugin_mcnp import *
from .plugin_serpent import *
from .plugin_abce import *
from .plugin_dakota import *
from .plugin_gcmat import *
from .results import *
from .template import *
from .parameters import *
from .database import *

# This allows a user to write watts.Quantity
from astropy.units import Quantity

__version__ = '0.5.2'



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

import csv
import json
import sys
import os
from pathlib import Path
import pickle
import subprocess
from typing import List, Optional

import numpy as np
import pandas as pd

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import PluginGeneric, _find_executable
from .results import Results, ExecInfo


class ResultsDakota(Results):
    """Dakota simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from Dakota run
    output_data
        Dictionary with data from .dat files
    """
    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        super().__init__(params, exec_info, inputs, outputs)
        self.output_data = self._get_Dakota_output(params)

    def _get_Dakota_output(self, params: Parameters) -> dict:
        """Read all Dakota '.dat' files and return results in a dictionary

        Returns
        -------
        Results from Dakota .dat files

        """

        dakota_out_file_name = params.get('dakota_out_file', 'dakota_opt.dat')

        # Save Dakota's main output '.dat' files
        output_data = {}

        if Path(dakota_out_file_name).exists():
            with open(dakota_out_file_name) as f:
                col_names = f.readline().split()
            df = pd.read_csv(dakota_out_file_name, sep=r"\s+", skiprows=1, names=col_names)

            for name in col_names:
                output_data[name] = np.array(df[name])

        # Save Dakota's final output '.dat' files
        if Path('finaldata1.dat').exists():
            with open('finaldata1.dat') as fd:
                reader = csv.reader(fd)
                rows = [row for idx, row in enumerate(reader) if idx == 0]

            new_rows = str(rows[0][0]).split()
            output_data['finaldata1'] = np.array([float(i) for i in new_rows])

        return output_data


class PluginDakota(PluginGeneric):
    """Plugin for running Dakota

    Parameters
    ----------
    template_file
        Templated Dakota input
    executable
        Path to Dakota script
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when Dakota is run
    show_stderr
        Whether to display output from stderr when Dakota is run

    Attributes
    ----------
    executable
        Path to DAKOTA script
    execute_command
        List of command-line arguments used to call the executable

    """
    def __init__(
        self,
        template_file: str,
        executable: PathLike = 'dakota.sh',
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        auto_link_files: Optional[str] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'DAKOTA_DIR')
        execute_command = ['{self.executable}', '-i', '{self.input_name}']
        super().__init__(
            executable, execute_command, template_file, extra_inputs,
            extra_template_inputs, "Dakota", show_stdout, show_stderr)

        self.input_name = template_file
        self._auto_link_files = auto_link_files

        # Setup to automatically include all 'extra_inputs' and 'extra_template_inputs'
        # to Dakota's "link files" option. Create a string of the file names in
        # 'extra_inputs' and 'extra_template_inputs'.
        if self._auto_link_files is not None:
            dakota_link_files = []
            if extra_inputs is not None:
                dakota_link_files.extend(list(extra_inputs))

            if extra_template_inputs is not None:
                dakota_link_files.extend(list(extra_template_inputs))

            self.dakota_link_files_string = " ".join(f"'{item}'" for item in dakota_link_files)

    def prerun(self, params: Parameters, filename: Optional[str] = None):
        """ Change the permisison of the Dakota driver file

        Parameters
        ----------
        params
            Parameters used to render template
        """

        # Store the string for the "link_files" in params.
        if self._auto_link_files is not None:
            params[self._auto_link_files] = self.dakota_link_files_string

        super().prerun(params)
        if 'dakota_driver_name' in params.keys():
            os.chmod(params['dakota_driver_name'], 0o755)

    @property
    def execute_command(self):
        return [str(self.executable), "-i", self.input_name]


def run_dakota_driver(coupled_code_exec: str):
    """ Function to execute the workflow for data
    exchange between Dakota and the coupled code
    in the Dakota driver script.

    Parameters
    ----------
    coupled_code_exec
        The name of the WATTS python script of the
        coupled code.
    """
    results = _parse_dakota_input()
    retval = _run_coupled_code(coupled_code_exec)
    _return_dakota_input(results, retval)


def _parse_dakota_input() -> Results:
    """Parse Dakota input

    Parameters
    ----------

    Returns
    -------
    results
        Results from Dakota's output file
    """

    from interfacing import interfacing as di # Dakota's interface module

    params, results = di.read_parameters_file()

    # Dump params to external params.json file for future use by the template engine
    params_for_template_engine_file_path = "params.json"
    with open(params_for_template_engine_file_path, 'w') as outfile:
        f = json.dump(params._variables,  outfile, default=lambda o: o.__dict__)
    return(results)


def _run_coupled_code(coupled_code_exec: str) -> dict:
    """ Run the coupled code

    Parameters
    ----------
    coupled_code_exec
        The name of the WATTS python script of the
        coupled code.

    Returns
    -------
    retval
        Processed output from the coupled code.
    """

    if not os.path.exists(coupled_code_exec):
        raise FileNotFoundError("Coupled-code script missing.")

    subprocess.check_output([sys.executable, coupled_code_exec])

    # Read the 'opt_res.out' pickle file and
    # store the results to 'res_output' for data
    # transfer with Dakota.
    if os.path.exists('opt_res.out'):
        db = pickle.load(open('opt_res.out', 'rb'))
        if 'dakota_descriptors' in db.keys():
            res_output = []
            for key in db['dakota_descriptors']:
                res_output.append(db[db['dakota_descriptors'][key]])
    else:
        raise RuntimeError("'opt_res.out' file is missing.")

    return {'fns': res_output}


def _return_dakota_input(results: Results, retval: dict):
    """ Return the output of the coupled code
    to Dakota.

    Parameters
    ----------
    results
        Results from Dakota's output file
    retval
        Processed output from the coupled code.
    """

    # Insert extracted values into results
    # Results iterator provides an index, response name, and response
    try:
        for i, n, r in results:
            if r.asv.function:
                # Returns the response function value from the coupled
                # code to Dakota (for pre Dakota v6.9).
                try:
                    r.function = retval['fns'][i]
                except:
                    pass
    # Catch Dakota 6.9 exception where results interface has changed
    # ValueError: too many values to unpack
    except ValueError:
        for i, (n, r) in enumerate(results.items()):
            r.function = retval['fns'][i]
    results.write()

    # Dump to external results.json file
    with open('results.json', 'w') as outfile:
        rst = json.dump(results, outfile, default=lambda o: o.__dict__)



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from functools import lru_cache
from pathlib import Path
from typing import Callable, Mapping, List, Optional

from uncertainties import ufloat

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import Plugin
from .results import Results, ExecInfo


class ResultsOpenMC(Results):
    """OpenMC simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, timestamp, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    keff
        K-effective value from the final statepoint
    statepoints
        List of statepoint files
    stdout
        Standard output from OpenMC run
    tallies
        List of OpenMC tally objects
    """

    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[Path], outputs: List[Path]):
        super().__init__(params, exec_info, inputs, outputs)

    @property
    def statepoints(self) -> List[Path]:
        return [p for p in self.outputs if p.name.startswith('statepoint')]

    @property
    @lru_cache()
    def keff(self) -> ufloat:
        import openmc
        # Get k-effective from last statepoint
        last_statepoint = self.statepoints[-1]
        with openmc.StatePoint(last_statepoint) as sp:
            if hasattr(sp, 'keff'):
                return sp.keff
            else:
                return sp.k_combined

    @property
    @lru_cache()
    def tallies(self) -> List:
        import openmc
        # Get k-effective from last statepoint
        last_statepoint = self.statepoints[-1]
        with openmc.StatePoint(last_statepoint) as sp:
            return list(sp.tallies.values())


class PluginOpenMC(Plugin):
    """Plugin for running OpenMC

    Parameters
    ----------
    model_builder
        Function that generates an OpenMC model
    extra_inputs
        Extra (non-templated) input files
    show_stdout
        Whether to display output from stdout when OpenMC is run
    show_stderr
        Whether to display output from stderr when OpenMC is run

    """

    def __init__(self, model_builder: Optional[Callable[[Parameters], None]] = None,
                 extra_inputs: Optional[List[PathLike]] = None,
                 show_stdout: bool = False, show_stderr: bool = False):
        super().__init__(extra_inputs, show_stdout, show_stderr)
        self.model_builder = model_builder
        self.unit_system = 'cgs'
        self.plugin_name = 'OpenMC'

    def prerun(self, params: Parameters) -> None:
        """Generate OpenMC input files

        Parameters
        ----------
        params
            Parameters used by the OpenMC template
        """
        # Convert quantities in parameters to CGS system
        params_copy = params.convert_units(system=self.unit_system)

        if self.model_builder is not None:
            self.model_builder(params_copy)

    def run(self, function: Optional[Callable] = None, **kwargs: Mapping):
        """Run OpenMC

        Parameters
        ----------
        function
            Function to execute. If not passed, defaults to only calling
            :func:`openmc.run`.
        **kwargs
            Keyword arguments passed on to ``function``

        See also
        --------
        openmc.run, openmc.plot_geometry, openmc.calculate_volumes

        """
        import openmc
        if function:
            function(**kwargs)
        else:
            openmc.run(**kwargs)

    def postrun(self, params: Parameters, exec_info: ExecInfo) -> ResultsOpenMC:
        """Collect information from OpenMC simulation and create results object

        Parameters
        ----------
        params
            Parameters used to create OpenMC model
        exec_info
            Execution information

        Returns
        -------
        OpenMC results object
        """

        def files_since(pattern, timestamp):
            matches = []
            for p in Path.cwd().glob(pattern):
                # Because of limited time resolution, we rely on access time to
                # determine input files
                mtime = p.stat().st_atime_ns
                if mtime >= timestamp:
                    matches.append(p)
            matches.sort(key=lambda x: x.stat().st_mtime_ns)
            return matches

        # Start with non-templated input files
        inputs = [Path.cwd() / p.name for p in self.extra_inputs]

        # Get generated input files
        for path in files_since('*.xml', exec_info.timestamp):
            if path not in inputs:
                inputs.append(path)

        # Get list of all output files
        outputs = ['OpenMC_log.txt']
        outputs.extend(files_since('tallies.out', exec_info.timestamp))
        outputs.extend(files_since('source.*.h5', exec_info.timestamp))
        outputs.extend(files_since('particle*.h5', exec_info.timestamp))
        outputs.extend(files_since('statepoint.*.h5', exec_info.timestamp))
        outputs.extend(files_since('volume*.h5', exec_info.timestamp))
        outputs.extend(files_since('*.png', exec_info.timestamp))
        outputs.extend(files_since('*.ppm', exec_info.timestamp))

        return ResultsOpenMC(params, exec_info, inputs, outputs)



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from collections import namedtuple
from datetime import datetime
from pathlib import Path
import shutil
from typing import List

import dill

from .fileutils import PathLike, open_file
from .parameters import Parameters


ExecInfo = namedtuple('ExecInfo', ['job_id', 'plugin', 'name', 'timestamp'])


class Results:
    """Results from running a workflow

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    base_path
        Path to directory storing results
    inputs
        List of input files
    job_id
        Integer ID of job
    outputs
        List of output files
    parameters
        Parameters used to generate inputs
    plugin
        Name of plugin
    stdout
        Standard output from execution
    time
        Time at which plugin was executed

    """

    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        self.base_path = Path.cwd()
        self.exec_info = exec_info
        self.parameters = Parameters(params)
        self.inputs = [Path(p) for p in inputs]
        self.outputs = [Path(p) for p in outputs]

    @property
    def plugin(self) -> str:
        return self.exec_info.plugin

    @property
    def time(self) -> datetime:
        return datetime.fromtimestamp(self.exec_info.timestamp * 1e-9)

    @property
    def job_id(self) -> int:
        return self.exec_info.job_id

    @property
    def name(self) -> str:
        return self.exec_info.name

    @property
    def stdout(self) -> str:
        return (self.base_path / f"{self.plugin}_log.txt").read_text()

    def move_files(self, dst: PathLike):
        """Move input/output files to different directory

        Parameters
        ----------
        dst
            Destination path where files should be moved

        """

        dst_path = Path(dst)
        # Move input/output files and change base -- note that trying to use the
        # Path.replace method doesn't work across filesystems, so instead we use
        # shutil.move
        for i, input in enumerate(self.inputs):
            shutil.move(str(input), str(dst_path / input.name))
            self.inputs[i] = dst_path / input.name
        for i, output in enumerate(self.outputs):
            shutil.move(str(output), str(dst_path / output.name))
            self.outputs[i] = dst_path / output.name
        self.base_path = dst_path

    def save(self, filename: PathLike):
        """Save results to a pickle file

        Parameters
        ----------
        filename
            File to save results to
        """
        with open(filename, 'wb') as fh:
            fh.write(dill.dumps(self))

    @classmethod
    def from_pickle(cls, filename: PathLike):
        """Load results from a pickle file

        Parameters
        ----------
        filename
            Path to load results from
        """
        with open(filename, 'rb') as fh:
            result =  dill.loads(fh.read())

        # For older results objects, add in execution info tuple
        if not hasattr(result, 'exec_info'):
            job_id = None
            if type(result) is Results:
                plugin = 'Generic'
            else:
                plugin = type(result).__name__[7:]
            name = result.__dict__['name']
            dt = result.__dict__['time']
            timestamp = int(dt.timestamp() * 1e6) * 1000
            result.exec_info = ExecInfo(job_id, plugin, name, timestamp)

        return result

    def open_folder(self):
        """Open folder containing results"""
        open_file(self.base_path)

    def __repr__(self):
        if self.name:
            return f"<Results: {self.plugin}, {self.name}, {self.time})>"
        else:
            return f"<Results: {self.plugin}, {self.time})>"




# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from __future__ import annotations
import copy
from collections import namedtuple
from collections.abc import MutableMapping, Mapping, Iterable
from datetime import datetime
from getpass import getuser
import textwrap
from typing import Any, Union, BinaryIO
from warnings import warn

import astropy.units as u
import dill
from prettytable import PrettyTable

# Enable imperial units
u.imperial.enable()

ParametersMetadata = namedtuple('ParametersMetadata', ['user', 'time'])


class Parameters(MutableMapping):
    """User parameters used to generate inputs that are created by plugins

    This class behaves like a normal Python dictionary except that it stores
    metadata on (key, value) pairs and provides the ability to save/load the
    data to a pickle file.

    Attributes
    ----------
    warn_duplicates : bool
        Whether to write a warning if a key is added and already exists

    """
    def __init__(self, *args, **kwargs):
        self._dict = {}
        self._metadata = {}
        self._warn_duplicates = False

        # Mimic the behavior of a normal dict object.
        if args:
            assert len(args) == 1
            args = args[0]
            if isinstance(args, Mapping):
                # dict(mapping)
                if hasattr(args, 'get_metadata'):
                    for key, value in args.items():
                        metadata = args.get_metadata(key)
                        self.set(key, value, **metadata._asdict())
                else:
                    for key, value in args.items():
                        self[key] = value
            elif isinstance(args, Iterable):
                # dict(iterable)
                for key, value in args:
                    self[key] = value
        elif kwargs:
            # dict(**kwargs)
            for key, value in kwargs.items():
                self[key] = value

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        self.set(key, value)

    def __delitem__(self, key):
        del self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return repr(self._dict)

    @property
    def warn_duplicates(self) -> bool:
        return self._warn_duplicates

    @warn_duplicates.setter
    def warn_duplicates(self, value):
        if not isinstance(value, bool):
            raise ValueError("warn_duplicates attribute must be a boolean.")
        self._warn_duplicates = value

    def set(self, key: Any, value: Any, *, user: str = None, time: datetime = None):
        """Explicitly set a key/value pair with metadata

        Parameters
        ----------
        key
            Key used in dictionary
        value
            Corresponding value in dictionary
        user
            Username associated with key/value pair
        time
            Time associated with key/value pair
        """
        if user is None:
            user = getuser()
        if time is None:
            time = datetime.now()
        if self._warn_duplicates and key in self._dict:
            warn(f"Key {key} has already been added to parameters")
        self._dict[key] = value
        self._metadata[key] = ParametersMetadata(user, time)

    def get_metadata(self, key: Any) -> ParametersMetadata:
        """Get metadata associated with a key

        Parameters
        ----------
        key
            Key to find metadata for

        Returns
        -------
        Associated metadata

        """
        return self._metadata[key]

    def show_summary(self, show_metadata=True, sort_by='key', filter_by={}):
        """Display a summary of the parameters

        Parameters
        ----------
        show_metadata
            Whether to print user and timestamp associated with key/value pair
        sort_by
            Field to sort by ('key', 'value', 'user', 'time'; default is 'key')
        filter_by
            Dictionary defining filters to apply to fields. The dictionary keys
            describe the fields and must be one of 'key', 'value', 'user', or
            'time'. The dictionary values are functions that will be applied to
            values in that field. A row will be displayed in the summary only
            if all functions applied to the values in that row return true.

        Examples
        --------
        Print all the parameters sorted by timestamp:

        >>> params.show_summary(sort_by='time')

        Print only the parameters added by <username>:

        >>> params.show_summary(filter_by={'user': lambda x: x == 'username'})
        """
        field_to_name = {
            'key': 'PARAMETER',
            'value': 'VALUE',
            'user': 'ADDED BY',
            'time': 'TIMESTAMP'
        }

        # Create a nicely-formatted table to display parameters
        headers = list(field_to_name.values())
        table = PrettyTable(field_names=headers, align='l')

        # Add rows to the table
        for key, value in self.items():
            md = self._metadata[key]

            # Apply the filters
            field_to_value = {
                'key': key, 'value': value, 'user': md.user, 'time': md.time}
            for field, filt in filter_by.items():
                if not filt(field_to_value[field]):
                    break
            else:
                table.add_row(
                    [key, textwrap.fill(str(value), width=40), md.user, md.time])

        # Sort and print summary
        if not show_metadata:
            headers = headers[:2]
        print(table.get_string(fields=headers, sortby=field_to_name[sort_by]))

    def _save_mapping(self, file_obj: BinaryIO):
        file_obj.write(dill.dumps(self))

    def save(self, filename_or_obj: Union[str, BinaryIO]):
        """Save parameters to a pickle file

        Parameters
        ----------
        filename_or_obj
            Path to open file or file object write to
        """
        if isinstance(filename_or_obj, str):
            with open(filename_or_obj, 'wb') as file_obj:
                self._save_mapping(file_obj)
        else:
            # If file object was passed, use it directly
            self._save_mapping(filename_or_obj)

    def _load_mapping(self, file_obj: BinaryIO):
        # Load parameters from pickle
        data = dill.loads(file_obj.read())

        # Update current instance
        self.update(data)
        self._metadata.update(data._metadata)

    def load(self, filename_or_obj: Union[str, BinaryIO]):
        """Load parameters from a pickle file

        Parameters
        ----------
        filename_or_obj
            Path to pickle file or file object to read from
        """
        if isinstance(filename_or_obj, str):
            with open(filename_or_obj, 'rb') as fh:
                self._load_mapping(fh)
        else:
            # If file object was passed, use it directly
            self._load_mapping(filename_or_obj)

    @classmethod
    def from_pickle(cls, filename_or_obj: Union[str, BinaryIO]) -> Parameters:
        """Return parameters from a pickle file

        Parameters
        ----------
        filename_or_obj
            Path to pickle file or file object to read from
        """
        params = cls()
        params.load(filename_or_obj)
        return params

    def convert_units(self, system: str = 'si', temperature: str = 'K',
                      inplace: bool = False) -> Parameters:
        """Perform unit conversion

        Parameters
        ----------
        system
            Desired unit system: 'si' or 'cgs'
        temperature
            Desired unit for temperature conversions
        inplace
            Whether to modify the parameters (True) or return a copy (False)

        Returns
        -------
        A :class:`Parameters` instance with converted units
        """
        params = self if inplace else copy.deepcopy(self)

        for key, value in params.items():
            if isinstance(value, u.Quantity):
                # Unit conversion for temperature needs to be done separately because
                # astropy uses a different method to convert temperature.
                if value.unit.physical_type == 'temperature':
                    params[key] = value.to(temperature, equivalencies=u.temperature()).value
                else:
                    params[key] = getattr(value, system).value

        return params



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from typing import List, Optional

import numpy as np
import pandas as pd

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import PluginGeneric, _find_executable
from .results import Results, ExecInfo


class ResultsMOOSE(Results):
    """MOOSE simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from MOOSE run
    csv_data
        Dictionary with data from .csv files
    """
    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        super().__init__(params, exec_info, inputs, outputs)
        self.csv_data = self._save_MOOSE_csv()

    def _save_MOOSE_csv(self) -> dict:
        """Read all MOOSE '.csv' files and return results in a dictionary

        Returns
        -------
        Results from MOOSE .csv files

        """
        input_file = self.inputs[0]
        csv_file = input_file.with_name(f"{input_file.stem}_csv.csv")

        # Save MOOSE's main output '.csv' files
        csv_data = {}
        if csv_file.exists():
            csv_file_df = pd.read_csv(csv_file)
            for column_name in csv_file_df.columns:
                csv_data[column_name] = np.array(csv_file_df[column_name])

        # Read MOOSE's vector postprocesssor '.csv' files and save the
        # parameters as individual array
        for output in self.outputs:
            if (output.name.startswith(f"{input_file.stem}_csv_") and
                not output.name.endswith("_0000.csv")):
                vector_csv_df = pd.read_csv(output)
                csv_param = list(set(vector_csv_df.columns) - {"id", "x", "y", "z"})
                csv_data[output.stem] = np.array(vector_csv_df[csv_param[0]], dtype=float)

                for name in ("id", "x", "y", "z"):
                    new_name = output.name[:-8] + name
                    if new_name not in csv_data:
                        csv_data[new_name] = np.array(vector_csv_df[name], dtype=float)

        return csv_data


class PluginMOOSE(PluginGeneric):
    """Plugin for running MOOSE

    Parameters
    ----------
    template_file
        Templated MOOSE input
    executable
        Path to MOOSE executable
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when MOOSE is run
    show_stderr
        Whether to display output from stderr when MOOSE is run

    Attributes
    ----------
    executable
        Path to MOOSE executable
    execute_command
        List of command-line arguments used to call the executable

    """

    def __init__(
        self,
        template_file: str,
        executable: PathLike = 'moose-opt',
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'MOOSE_DIR')
        execute_command = ['{self.executable}', '-i', '{self.input_name}']
        super().__init__(
            executable, execute_command, template_file, extra_inputs,
            extra_template_inputs, "MOOSE", show_stdout, show_stderr)
        self.input_name = "MOOSE.i"



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

from pathlib import Path
import shutil
import subprocess
from typing import List, Optional

import numpy as np
import pandas as pd

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import PluginGeneric, _find_executable
from .results import Results, ExecInfo


class ResultsRELAP5(Results):
    """RELAP5 simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from RELAP5 run
    csv_data
        Dictionary with data from .csv files
    """
    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        super().__init__(params, exec_info, inputs, outputs)
        self.csv_data = self._get_relap5_csv_data()

    def _get_relap5_csv_data(self) -> dict:
        """Read relap5 '.csv' file and return results in a dictionary

        Returns
        -------
        Results from relap5 .csv files

        """
        csv_data = {}
        if Path('R5-out.csv').exists():
            csv_file_df = pd.read_csv('R5-out.csv')
            for column_name in csv_file_df.columns:
                csv_data[column_name] =  np.array(csv_file_df[column_name])
        return csv_data

class PluginRELAP5(PluginGeneric):
    """Plugin for running RELAP5

    Parameters
    ----------
    template_file
        Templated RELAP5 input
    executable
        Path to RELAP5 executable
    plotfl_to_csv
        Whether to convert RELAP5's plotfl file to CSV file
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when RELAP5 is run
    show_stderr
        Whether to display output from stderr when RELAP5 is run

    """

    def  __init__(
        self,
        template_file: str,
        executable: PathLike = 'relap5.x',
        plotfl_to_csv: bool = True,
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'RELAP5_DIR')
        execute_command = ['{self.executable}', '-i', '{self.input_name}']
        super().__init__(
            executable, execute_command, template_file, extra_inputs,
            extra_template_inputs, "RELAP5", show_stdout, show_stderr)
        self.input_name = "RELAP5.i"
        self.plotfl_to_csv = plotfl_to_csv

    def run(self, extra_args: Optional[List[str]] = None):
        """Run RELAP5

        Parameters
        ----------
        extra_args
            Extra arguments to be appended when running the RELAP5 executable
        """

        # Copy all necessary files to the temporary directory. RELAP5 requires
        # the executable file and the license key to be in the same directory as
        # the input file to run. Users can also add all fluid property files
        # here.
        cwd = Path.cwd()
        for fname in self.executable.parent.iterdir():
            shutil.copy2(str(fname), str(cwd))

        # Create a list for RELAP5 input command and append any extra
        # options to it.
        if extra_args is None:
            extra_args = []
        command = self.execute_command + extra_args

        # run_proc() does not work with RELAP5-3D.
        # The extra argument of 'stdout' to subprocess.Popen() in run_proc() somehow prevents RELAP5 from running.
        # As a work-around, we explicitly use subprocess.Popen() here without specifying 'stdout=subprocess.PIPE')
        p = subprocess.Popen(command)
        stdout, stderr = p.communicate()

    def postrun(self, params: Parameters, exec_info: ExecInfo) -> ResultsRELAP5:
        """Read RELAP5 results and create results object

        Parameters
        ----------
        params
            Parameters used to create RELAP5 model
        exec_info
            Execution information

        Returns
        -------
        RELAP5 results object
        """

        # Convert RELAP5's plotfl file to CSV file for processing
        if self.plotfl_to_csv:
            if Path('plotfl').exists():
                self._plotfl_to_csv()
            else:
                raise RuntimeError(
                    "Output plot file 'plotfl' is missing. Please make sure you "
                    "are running the correct version of RELAP5-3D or the plot "
                    "file is named correctly."
                )

        return super().postrun(params, exec_info)

    # The RELAP5-3D version used here does not generate csv output files.
    # It generates a text file with a particular format that needs to
    # be converted to a csv file before the results can be extracted.

    def _plotfl_to_csv(self):
        """Converts RELAP5's plotfl file to csv.

        """

        with open('plotfl') as f:
            contents = f.readlines()

        # Get markers for 'contents'
        n_plotinf = self._check_string(contents, 'plotinf')
        n_plotalf = self._check_string(contents, 'plotalf')
        n_plotnum = self._check_string(contents, 'plotnum')
        n_plotrec = self._check_string(contents, 'plotrec')

        # Break 'contents' into 3 sections: channel, ID, values
        channels = contents[n_plotalf[0] : n_plotnum[0]]
        ids = contents[n_plotnum[0] : n_plotrec[0]]
        values = contents[n_plotrec[0] : ]

        # Extract values
        channel_list = self._extract_value(channels)
        id_list = self._extract_value(ids)

        # Create a dictionary to store data
        data_dict = {'channel': channel_list, 'id': id_list}

        for i in range(len(n_plotrec)):
            i_start = n_plotrec[i] - n_plotrec[0] # Subtract n_plotrec[0] to offset the the 'content' above 'values'
            i_end = n_plotrec[i] - n_plotrec[0] + (n_plotrec[1] - n_plotrec[0]) # Add "n_plotrec[1] - n_plotrec[0]" to reach the last line of 'values'
            value_list = np.double(self._extract_value(values[i_start : i_end]))
            data_dict[f"value_t_{i}"] = value_list

        # Convert the dictionary into DataFrame
        df = pd.DataFrame(data_dict)
        df.set_index('channel', inplace=True)
        df = df.T

        # Create new column for DataFrame
        new_col = []
        for i in range(len(df.columns)):
            new_col.append(f"{df.columns[i]}-{df.iloc[0][i]}")

        df.columns = new_col
        df = df[1:] # Remove the volume ID row

        # Save DataFrame as csv
        df.to_csv('R5-out.csv')

    def _check_string(self, file, keyword: str) -> list:
        """Looks for certain keyword markers in the plotfl file.

        Parameters
        ----------
        file
            RELAP5 output file
        keyword
            Keyword whose location to look for
        Returns
        -------
        A list of line number of the keyword markers in the plotfl file

        """
        return [i for i, file_i in enumerate(file) if keyword in file_i]

    def _extract_value(self, contents):
        """Extracts values from the plotfl file according to the keyword markers.

        Parameters
        ----------
        contents
            RELAP5 output file broken up according to the keyword markers
        Returns
        -------
        List of extracted values.

        """
        value_list = []
        s = ''
        for line in contents:
            value_list.append(s.strip())
            s = ''
            for char_val in line:
                if char_val != ' ':
                    s += str(char_val)
                else:
                    value_list.append(s.strip())
                    s = ''
        value_list.append(s.strip())

        # Remove all spaces
        while '' in value_list:
            value_list.remove('')

        # Remove first element, i.e. markers
        del value_list[0]

        return value_list



# SPDX-FileCopyrightText: 2022-2025 UChicago Argonne, LLC
# SPDX-License-Identifier: MIT

import glob
import os
from pathlib import Path
import shutil
import subprocess
from typing import List, Optional

import numpy as np
import pandas as pd

from .fileutils import PathLike
from .parameters import Parameters
from .plugin import PluginGeneric, _find_executable
from .results import Results, ExecInfo


class ResultsSAS(Results):
    """SAS simulation results

    Parameters
    ----------
    params
        Parameters used to generate inputs
    exec_info
        Execution information (job ID, plugin name, time, etc.)
    inputs
        List of input files
    outputs
        List of output files

    Attributes
    ----------
    stdout
        Standard output from SAS run
    csv_data
        Dictionary with data from .csv files
    """
    def __init__(self, params: Parameters, exec_info: ExecInfo,
                 inputs: List[PathLike], outputs: List[PathLike]):
        super().__init__(params, exec_info, inputs, outputs)
        self.csv_data = self._get_sas_csv_data()

    def _get_sas_csv_data(self) -> dict:
        """Read all sas '.csv' files and return results in a dictionary

        Returns
        -------
        Results from sas .csv files

        """
        csv_files = glob.glob("*.csv") # List all csv files

        csv_data = {}
        for file in csv_files:
            if os.path.getsize(file) > 0: # Check if file is empty
                csv_data_sub = {}
                csv_file_df = pd.read_csv(file)

                for column_name in csv_file_df.columns:
                    csv_data_sub[column_name] =  np.array(csv_file_df[column_name])

                csv_data[file[:-4]] = csv_data_sub # Save sub dictionary as csv file name
                                                   # Removed .csv extension from file name

        return csv_data


class PluginSAS(PluginGeneric):
    """Plugin for running SAS

    Parameters
    ----------
    template_file
        Templated SAS input
    executable
        Path to SAS executable
    extra_inputs
        List of extra (non-templated) input files that are needed
    extra_template_inputs
        Extra templated input files
    show_stdout
        Whether to display output from stdout when SAS is run
    show_stderr
        Whether to display output from stderr when SAS is run

    Attributes
    ----------
    executable
        Path to SAS executable
    execute_command
        List of command-line arguments used to call the executable
    conv_channel
        Path to CHANNELtoCSV utility executable
    conv_mfuel
        Path to MFUELPeaktoCSV utility executable
    conv_primar4
        Path to PRIMAR4toCSV utility executable

    """

    def  __init__(
        self,
        template_file: str,
        executable: PathLike = 'sas.x',
        extra_inputs: Optional[List[str]] = None,
        extra_template_inputs: Optional[List[PathLike]] = None,
        show_stdout: bool = False,
        show_stderr: bool = False
    ):
        executable = _find_executable(executable, 'SAS_DIR')
        execute_command = ['{self.executable}', '-i', '{self.input_name}',
                           '-o', 'out.txt']
        super().__init__(executable, execute_command, template_file, extra_inputs,
                         extra_template_inputs, "SAS", show_stdout, show_stderr)
        self.input_name = "SAS.inp"

        # Set other executables based on the main SAS executable
        suffix = executable.suffix
        self._conv_channel = executable.with_name(f"CHANNELtoCSV{suffix}")
        self._conv_mfuel = executable.with_name(f"MFUELtoCSV{suffix}")
        self._conv_primar4 = executable.with_name(f"PRIMAR4toCSV{suffix}")

    @property
    def conv_channel(self) -> Path:
        return self._conv_channel

    @property
    def conv_primar4(self) -> Path:
        return self._conv_primar4

    @property
    def conv_mfuel(self) -> Path:
        return self._conv_mfuel

    @conv_channel.setter
    def conv_channel(self, exe: PathLike):
        if shutil.which(exe) is None:
            raise RuntimeError(f"CHANNELtoCSV utility executable '{exe}' is missing.")
        self._conv_channel = Path(exe)

    @conv_primar4.setter
    def conv_primar4(self, exe: PathLike):
        if shutil.which(exe) is None:
            raise RuntimeError(f"PRIMAR4toCSV utility executable '{exe}' is missing.")
        self._conv_primar4 = Path(exe)

    @conv_mfuel.setter
    def conv_mfuel(self, exe: PathLike):
        if shutil.which(exe) is None:
            raise RuntimeError(f"MFUELtoCSV utility executable '{exe}' is missing.")
        self._conv_mfuel = Path(exe)

    @property
    def execute_command(self):
        return [str(self.executable), "-i", self.input_name, "-o", "out.txt"]

    def postrun(self, params: Parameters, exec_info: ExecInfo) -> ResultsSAS:
        """Read SAS results and create results object

        Parameters
        ----------
        params
            Parameters used to create SAS model
        name
            Name of the workflow

        Returnss
        -------
        SAS results object
        """

        # Convert CHANNEl.dat and PRIMAR4.dat and MFUELss.dat to csv files
        # using SAS utilities. Check if files exist because
        # they may not be outputted per user's choice.

        if Path("CHANNEL.dat").is_file():
            with open("CHANNEL.dat", "r") as file_in, open("CHANNEL.csv", "w") as file_out:
                subprocess.run(str(self.conv_channel), stdin=file_in, stdout=file_out)

        if Path("PRIMAR4.dat").is_file():
            with open("PRIMAR4.dat", "r") as file_in, open("PRIMAR4.csv", "w") as file_out:
                subprocess.run(str(self.conv_primar4), stdin=file_in, stdout=file_out)

        if Path("MFUELss_C000001.dat").is_file():
            with open("MFUELss_C000001.dat", "r") as file_in:
                subprocess.run(str(self.conv_mfuel), stdin=file_in)

        return super().postrun(params, exec_info)
